import 'dart:io';

import 'package:dio/dio.dart';
import 'package:flutter/cupertino.dart';
import 'package:hgg_base/hgg_base.dart';
import 'package:hgg_base/src/utils/x_log.dart';
import 'package:hgg_base/src/utils/x_util.dart';
import 'package:pull_to_refresh/pull_to_refresh.dart';

/// 页面状态
enum PageState {
  Normal, //正常
  Loading, //加载中
  Empty, //空数据
  Error, //错误
}

/// 基础 model
abstract class XBaseModel extends GetxController {
  /// 获取 model
  S? findModel<S>({String? tag}) {
    try {
      return GetInstance().find<S>(tag: tag);
    } catch (e) {
      XLog.e(e);
      return null;
    }
  }

  /// 当前的页面状态,默认为normal
  final _pageState = PageState.Loading.obs;

  PageState get pageState => _pageState.value;

  bool get normal => _pageState.value == PageState.Normal;

  bool get loading => _pageState.value == PageState.Loading;

  bool get empty => _pageState.value == PageState.Empty;

  bool get error => _pageState.value == PageState.Error;

  /// 出错时的message
  String _errorMessage = '加载失败';

  String get errorMessage => _errorMessage;

  final _buttonLoading = false.obs;

  bool get buttonLoading => _buttonLoading.value; // 按钮加载状态

  void doLoading() => _buttonLoading.value = true;

  void endLoading() => _buttonLoading.value = false;

  /// 当前 model,控制输入框 node
  late List<FocusNode> nodeList;

  int get nodeNum => 0; //node数量

  /// 关闭键盘
  void hideKeyboard() {
    nodeList.forEach((node) {
      if (node.hasFocus) node.unfocus();
    });
  }

  @override
  void onInit() {
    super.onInit();
    if (nodeNum > 0) {
      nodeList = List.generate(nodeNum, (index) => FocusNode());
    } else {
      nodeList = [];
    }
  }

  /// 统一处理异常情况
  /// [e],有可能是Error,也有可能是Exception.所以需要判断处理
  /// [s] 为堆栈信息
  void _handleCatch(e, s) {
    if (e is DioError) {
      if (e.error is SocketException) {
        _setError('网络错误');
      } else if (e.type == DioErrorType.connectTimeout) {
        _setError('连接超时,请检测网络是否正常');
      } else {
        final requestUrl = Uri.decodeComponent(e.requestOptions.uri.toString());
        XLog.e(e.toString(), 'DioError错误url:$requestUrl');
        _setError(e.message);
      }
    } else {
      XLog.e(e.toString(), '网络请求,非 DioError 错误');
      XLog.e(s.toString(), '非 DioError 错误,堆栈信息');

      _setError(e is Error ? e.toString() : e.message);
    }
  }

  /// 设置错误状态
  void _setError(String message) {
    if (message.contains('[500]')) message = '服务升级中,请稍等';
    if (message.length < 50) {
      _errorMessage = message;
    }
    _pageState.value = PageState.Error;
  }

  /// 初始化数据
  refreshWithLoading() {}

  /// 获取数据前执行
  Future beforeGetData() async {}
}

/// 初始请求数据的 Model
abstract class XDataModel<T> extends XBaseModel {
  //获取数据
  Rxn<T> _data = Rxn<T>();

  T? get data => _data.value;

  /// 更新 data 内数据
  updateData(void fn(T? val)) {
    _data.update(fn);
  }

  @override
  void onInit() {
    super.onInit();
    _initData();
  }

  /// 获取 Sp 中缓存的数据
  T? getSpData() => null;

  /// 初始化数据
  _initData() async {
    await beforeGetData();
    if (getSpData() != null) {
      // 有缓存数据先加载缓存数据
      _data.value = getSpData();
      // 获取到缓存后,继续请求网络,已获取最新的数据
      await refreshData();
    } else {
      // 未获取到缓存数据,直接请求网络获取
      await refreshWithLoading();
    }
  }

  @override
  refreshWithLoading() async {
    _pageState.value = PageState.Loading;
    await refreshData();
  }

  /// 刷新数据-从网络获取
  Future<void> refreshData() async {
    try {
      _data.value = await getData();
      afterGetData();
      _pageState.value = PageState.Normal;
    } catch (e, s) {
      _handleCatch(e, s);
    }
  }

  /// 获取初始化加载数据
  Future<T> getData();

  /// 获取数据后
  void afterGetData() {}
}

/// 初始请求List的 Model
abstract class XListModel<T> extends XBaseModel {
  //获取数据
  final _list = <T>[].obs;

  RxList<T> get list => _list;

  /// 分页第一页页码
  late int _pageNumFirst;

  int get pageNumFirst => _pageNumFirst;

  /// 分页条目数量
  late int _pageSize;

  RefreshController _refreshController =
      RefreshController(initialRefresh: false);

  RefreshController get refreshController => _refreshController;

  /// 当前页码
  late int _currentPageNum;

  int get currentPageNum => _currentPageNum;

  @override
  void onInit() {
    super.onInit();
    // _refreshController = RefreshController(initialRefresh: false);
    _initList();
  }

  @override
  void onClose() {
    _refreshController.dispose();
    super.onClose();
  }

  int initFirstPageNum() => 0;

  int initPageSize() => 15;

  /// 初始化数据
  _initList() async {
    _pageNumFirst = initFirstPageNum();
    _pageSize = initPageSize();

    await beforeGetData();
    if (getSpData() != null) {
      // 有缓存数据先加载缓存数据
      final data = getSpData();
      if (data == null || data.isEmpty) {
        _pageState.value = PageState.Empty;
      } else {
        _list.assignAll(data);
      }
      // 获取到缓存后,继续请求网络,已获取最新的数据
      await refreshData();
    } else {
      // 未获取到缓存数据,直接请求网络获取
      await refreshWithLoading();
    }
  }

  @override
  refreshWithLoading() async {
    _pageState.value = PageState.Loading;
    await refreshData();
  }

  /// 获取 Sp 中缓存的数据
  List<T>? getSpData() => null;

  /// 刷新数据
  Future<void> refreshData() async {
    _currentPageNum = _pageNumFirst;
    try {
      final data = (await getList(pageNum: _pageNumFirst)) ?? [];
      if (U.isNullOrEmpty(data)) {
        _list.assignAll([]);
        afterGetList(data, _pageNumFirst);
        _pageState.value = PageState.Empty;
      } else {
        _list.assignAll(data);
        refreshController.refreshCompleted();

        if (data.length < _pageSize) {
          if (_currentPageNum != _pageNumFirst) {
            refreshController.loadNoData();
          } else {
            refreshController.loadComplete();
          }
        } else {
          //防止上次上拉加载更多失败,需要重置状态
          refreshController.loadComplete();
        }

        /// 获取数据后
        afterGetList(data, _pageNumFirst);
        _pageState.value = PageState.Normal;
      }
    } catch (e, s) {
      _handleCatch(e, s);
    }
  }

  /// 获取初始化加载数据
  Future<List<T>?> getList({required int pageNum});

  /// 更新 页面状态
  refreshPageState() {
    if (_list.isEmpty) {
      _pageState.value = PageState.Empty;
    } else {
      _pageState.value = PageState.Normal;
    }
  }

  /// 获取数据后
  void afterGetList(List<T> list, int pageNum) {}

  /// 上拉加载更多
  Future<List<T>> loadMore() async {
    try {
      final data = (await getList(pageNum: ++_currentPageNum)) ?? [];
      if (data.isEmpty) {
        _currentPageNum--;
        refreshController.loadNoData();
      } else {
        _list.addAll(data);
        if (data.length < _pageSize) {
          refreshController.loadNoData();
        } else {
          refreshController.loadComplete();
        }

        /// 获取数据后
        afterGetList(data, currentPageNum);
      }
      return data;
    } catch (e, s) {
      _currentPageNum--;
      refreshController.loadFailed();
      XLog.e('error--->\n' + e.toString());
      XLog.e('statck--->\n' + s.toString());
      return [];
    }
  }

  /// 下拉加载更多(消息中)
  Future<List<T>> messageLoadMore() async {
    try {
      final data = (await getList(pageNum: ++_currentPageNum)) ?? [];
      if (data.isEmpty) {
        _currentPageNum--;
        refreshController.loadNoData();
      } else {
        //将新列表放在列表的前面
        _list.insertAll(0, data);
        if (data.length < _pageSize) {
          refreshController.loadNoData();
        } else {
          refreshController.loadComplete();
        }

        /// 获取数据后
        afterGetList(data, currentPageNum);
      }
      return data;
    } catch (e, s) {
      _currentPageNum--;
      refreshController.loadFailed();
      XLog.e('error--->\n' + e.toString());
      XLog.e('statck--->\n' + s.toString());
      return [];
    }
  }
}
